# Generated by Django 2.0.13 on 2019-05-17 07:54

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='CMISConfig',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('client_url', models.URLField(default='http://localhost:8082/alfresco/cmisatom')),
                ('client_user', models.CharField(default='admin', max_length=200)),
                ('client_password', models.CharField(default='admin', max_length=200)),
            ],
            options={
                'verbose_name': 'CMIS Configuration',
            },
        ),
    ]



from django.contrib import admin

from solo.admin import SingletonModelAdmin

from .models import CMISConfig, CMISFolderLocation


@admin.register(CMISConfig)
class CMISConfigAdmin(SingletonModelAdmin):
    filter_horizontal = ['locations']


@admin.register(CMISFolderLocation)
class CMISFolderLocationAdmin(admin.ModelAdmin):
    pass



from django.apps import AppConfig
from django.core.checks import Error, Tags, register


class CMISConfig(AppConfig):
    name = "drc_cmis"
    verbose_name = "CMIS"
    app_name = "cmis"

    def ready(self):
        # pass
        register(check_cmis, Tags.compatibility, deploy=True)


def check_cmis(app_configs, **kwargs):
    """
    ZDS 1.2.01, hoofdstuk 5:
    Ten behoeve van de integratie met het DRC en het vastleggen van
    zaakdocumenten dient het DMS aan de volgende eisen te voldoen:

        * Het DMS wordt ontsloten als een CMIS 1.0 repository;
        * De CMIS-interface dient minimaal navolgende opties te ondersteunen:
            - "Multi-filing";
            - "Change Log", met registratie van Change Events voor
              filing/unfiling/moving van de objecten documenten en folders;
            - Nieuwe CMIS-objecttypes van het Base Type "cmis:document" en
              "cmis:folder" worden ondersteund;
        * De CMIS-changelog is toegankelijk voor het DRC.

    :param app_configs:
    :param kwargs:
    :return:
    """
    from .client import cmis_client
    from .choices import CMISCapabilities

    errors = []
    try:
        capabilities = cmis_client._repo.capabilities
    except Exception:
        errors.append(
            Error(
                "Could not communicate with the DMS.",
                hint="Make sure the authentication and host settings are correct.",
            )
        )
        return errors
    else:
        multifiling = capabilities.get(CMISCapabilities.multifiling, None)
        if not multifiling:
            errors.append(Error("The DMS does not support Multifiling, or it's disabled."))
        unfiling = capabilities.get(CMISCapabilities.unfiling, None)
        if not unfiling:
            errors.append(Error("The DMS does not support Unfiling or it's disabled."))
        changes = capabilities.get(CMISCapabilities.changes, None)
        if not changes or changes == "none":
            errors.append(
                Error(
                    "The DMS does not support Change Log, or it's disabled.",
                    hint="In case you're running Alfresco, make sure to add the relevant audit.* properties.",
                )
            )

    return errors



# Generated by Django 2.0.13 on 2019-05-23 10:54

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('drc_cmis', '0001_initial_new'),
    ]

    operations = [
        migrations.CreateModel(
            name='CMISFolderLocation',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('location', models.CharField(default='', max_length=200)),
            ],
        ),
        migrations.AddField(
            model_name='cmisconfig',
            name='locations',
            field=models.ManyToManyField(to='drc_cmis.CMISFolderLocation'),
        ),
    ]



MAP = {
    # "uuid": "drc:",
    "identificatie": "drc:document__identificatie",
    "bronorganisatie": "drc:document__bronorganisatie",
    "creatiedatum": "drc:document__creatiedatum",
    "titel": "cmis:name",
    "vertrouwelijkheidaanduiding": "drc:document__vertrouwelijkaanduiding",
    "auteur": "drc:document__auteur",
    "status": "drc:document__status",
    "beschrijving": "drc:document__beschrijving",
    "ontvangstdatum": "drc:document__ontvangstdatum",
    "verzenddatum": "drc:document__verzenddatum",
    "indicatie_gebruiksrecht": "drc:document__indicatiegebruiksrecht",
    "ondertekening_soort": "drc:document__ondertekeningsoort",
    "ondertekening_datum": "drc:document__ondertekeningdatum",
    "informatieobjecttype": "drc:document__informatieobjecttype",
    "formaat": "drc:document__formaat",
    "taal": "drc:document__taal",
    "bestandsnaam": "drc:document__bestandsnaam",
    # "inhoud": "drc:",
    "link": "drc:document__link",
    "integriteit_algoritme": "drc:document__integriteitalgoritme",
    "integriteit_waarde": "drc:document__integriteitwaarde",
    "integriteit_datum": "drc:document__integriteitdatum",
}

REVERSE_MAP = {
    value: key for key, value in MAP.items()
}

def mapper(drc_name):
    return MAP.get(drc_name, None)


def reverse_mapper(cmis_name):
    return REVERSE_MAP.get(cmis_name, None)



import base64

from django.conf import settings
from django.urls import reverse


def create_dict_from_cmis_doc(cmis_doc):
    properties = cmis_doc.getProperties()

    # Values that need some parsing.
    creatiedatum = properties.get("drc:document__creatiedatum")
    if creatiedatum:
        creatiedatum = creatiedatum.date()

    ontvangstdatum = properties.get("drc:document__ontvangstdatum")
    if ontvangstdatum:
        ontvangstdatum = ontvangstdatum.date()

    verzenddatum = properties.get("drc:document__verzenddatum")
    if verzenddatum:
        verzenddatum = verzenddatum.date()

    ondertekening_datum = properties.get("drc:document__ondertekeningdatum")
    if ondertekening_datum:
        ondertekening_datum = ondertekening_datum.date()

    integriteit_datum = properties.get("drc:document__integriteitdatum")
    if integriteit_datum:
        integriteit_datum = integriteit_datum.date()

    identificatie = properties.get("drc:document__identificatie")

    cmis_id = properties.get("cmis:versionSeriesId").split('/')[-1]

    try:
        inhoud = base64.b64encode(cmis_doc.getContentStream().read()).decode("utf-8")
    except AssertionError:
        return None
    else:
        url = "{}{}".format(settings.HOST_URL, reverse(
            'enkelvoudiginformatieobjecten-detail', kwargs={'version': '1', 'uuid': cmis_id}
        ))
        return {
            "url": url,
            "inhoud": url,
            "creatiedatum": creatiedatum,
            "ontvangstdatum": ontvangstdatum,
            "verzenddatum": verzenddatum,
            "integriteit_datum": integriteit_datum,
            "ondertekening_datum": ondertekening_datum,
            "titel": properties.get("cmis:name"),
            "identificatie": identificatie,
            "bronorganisatie": properties.get("drc:document__bronorganisatie"),
            "vertrouwelijkaanduiding": properties.get("drc:document__vertrouwelijkaanduiding"),
            "auteur": properties.get("drc:document__auteur"),
            "status": properties.get("drc:document__status"),
            "beschrijving": properties.get("drc:document__beschrijving"),
            "indicatie_gebruiksrecht": properties.get("drc:document__indicatiegebruiksrecht"),
            "ondertekening_soort": properties.get("drc:document__ondertekeningsoort"),
            "informatieobjecttype": properties.get("drc:document__informatieobjecttype"),
            "formaat": properties.get("drc:document__formaat"),
            "taal": properties.get("drc:document__taal"),
            "bestandsnaam": properties.get("drc:document__bestandsnaam"),
            "link": properties.get("drc:document__link"),
            "integriteit_algoritme": properties.get("drc:document__integriteitalgoritme"),
            "integriteit_waarde": properties.get("drc:document__integriteitwaarde"),
            "bestandsomvang": len(inhoud),
        }


def create_case_dict_from_cmis_doc(cmis_doc):
    properties = cmis_doc.getProperties()

    cmis_id = properties.get("cmis:versionSeriesId").split('/')[-1]
    url = "{}{}".format(settings.HOST_URL, reverse(
        'objectinformatieobjecten-detail', kwargs={'version': '1', 'uuid': cmis_id}
    ))
    eio_url = "{}{}".format(settings.HOST_URL, reverse(
        'enkelvoudiginformatieobjecten-detail', kwargs={'version': '1', 'uuid': cmis_id}
    ))

    return {
        "url": url,
        "informatieobject": eio_url,
        "object": properties.get('drc:connectie__zaakurl'),
        "object_type": properties.get("drc:connectie__objecttype"),
        "aard_relatie_weergave": properties.get("drc:connectie__aardrelatieweergave"),
        "titel": properties.get("drc:connectie__titel"),
        "beschrijving": properties.get("drc:connectie__beschrijving"),
        "registratiedatum": properties.get("drc:connectie__registratiedatum"),
    }



class DMSException(Exception):
    def __init__(self, message, *args, **kwargs):
        self.message = message


class DocumentExistsError(DMSException):
    pass


class DocumentDoesNotExistError(DMSException):
    pass


class SyncException(DMSException):
    pass


class DocumentConflictException(DMSException):
    pass


class DocumentLockedException(DMSException):
    pass



from rest_framework.exceptions import ValidationError


class BackendException(ValidationError):
    def __init__(self, detail, create=False, update=False, retreive_single=False, delete=False, retreive_list=False, code=None):
        self.create = create
        self.update = update
        self.retreive_single = retreive_single
        self.delete = delete
        self.retreive_list = retreive_list

        super().__init__(detail, code)


class AbstractStorageBackend:
    exception_class = BackendException



from io import BytesIO

from django.test import TestCase

from rest_framework.exceptions import ErrorDetail
from tests.app.backend import BackendException

from drc_cmis import settings
from drc_cmis.backend import CMISDRCStorageBackend
from drc_cmis.client import cmis_client
from drc_cmis.models import CMISConfig, CMISFolderLocation

from .factories import EnkelvoudigInformatieObjectFactory
from .mixins import DMSMixin


class CMISStorageTests(DMSMixin, TestCase):
    def setUp(self):
        super().setUp()

        self.backend = CMISDRCStorageBackend()
        location = CMISFolderLocation.objects.create(location=settings.BASE_FOLDER_LOCATION)
        config = CMISConfig.get_solo()
        config.locations.add(location)

    # CREATE DOCUMENT TESTS
    def test_create_document(self):
        eio = EnkelvoudigInformatieObjectFactory()
        document = self.backend.create_document(eio.__dict__.copy(), BytesIO(b'some content'))
        self.assertIsNotNone(document)

    def test_create_document_error_update_conflict(self):
        eio = EnkelvoudigInformatieObjectFactory()
        eio_dict = eio.__dict__

        document = self.backend.create_document(eio_dict.copy(), BytesIO(b'some content'))
        self.assertIsNotNone(document)

        eio_dict['identificatie'] = 'test'

        with self.assertRaises(BackendException) as exception:
            self.backend.create_document(eio_dict.copy(), BytesIO(b'some content'))
        self.assertEqual(exception.exception.detail, {None: ErrorDetail(string='Document is niet uniek. Dit kan liggen aan de titel, inhoud of documentnaam', code='invalid')})

    def test_create_document_error_identification_exists(self):
        eio = EnkelvoudigInformatieObjectFactory()
        eio_dict = eio.__dict__
        eio_dict['identificatie'] = 'test'

        document = self.backend.create_document(eio_dict.copy(), BytesIO(b'some content'))
        self.assertIsNotNone(document)

        eio_dict['titel'] = 'gewoon_een_andere_titel'
        eio_dict['identificatie'] = 'test'

        with self.assertRaises(BackendException) as exception:
            self.backend.create_document(eio_dict.copy(), BytesIO(b'some content'))
        self.assertEqual(exception.exception.detail, {None: ErrorDetail(string=f"Document identificatie {eio.identificatie} is niet uniek", code='invalid')})

    # GET DOCUMENT TESTS
    def test_get_documents(self):
        eio = EnkelvoudigInformatieObjectFactory()
        doc = self.backend.create_document(eio.__dict__, BytesIO(b'some content'))
        self.assertIsNotNone(doc)

        # Because we need to give alfresco some time to index the document there is a timeout
        import time
        time.sleep(25)

        documents = self.backend.get_documents()
        self.assertEqual(len(documents), 1)

    def test_get_documents_empty(self):
        documents = self.backend.get_documents()

        self.assertEqual(len(documents), 0)

    def test_get_document(self):
        eio = EnkelvoudigInformatieObjectFactory(identificatie='test')
        eio_dict = eio.__dict__

        document = self.backend.create_document(eio_dict.copy(), None)
        self.assertIsNotNone(document)
        self.assertEqual(document['identificatie'], 'test')

        fetched_document = self.backend.get_document(document['url'].split('/')[-1])
        self.assertEqual(document['titel'], fetched_document['titel'])

    # TODO: Fix this call
    # def test_get_document_cases(self):
    #     eio = EnkelvoudigInformatieObjectFactory(identificatie='test')
    #     eio_dict = eio.__dict__

    #     document = self.backend.create_document(eio_dict.copy(), None)
    #     self.assertIsNotNone(document)
    #     self.assertEqual(document['identificatie'], 'test')

    #     documents = self.backend.get_document_cases()
    #     self.assertGreaterEqual(len(documents), 1)

    # UPDATE DOCUMENT TESTS
    def test_update_document(self):
        eio = EnkelvoudigInformatieObjectFactory(identificatie='test')
        eio_dict = eio.__dict__

        document = self.backend.create_document(eio_dict.copy(), BytesIO(b'some content'))
        self.assertIsNotNone(document)
        self.assertEqual(document['identificatie'], 'test')

        eio_dict['titel'] = 'test-titel-die-unique-is'

        updated_document = self.backend.update_document(eio_dict.copy(), eio_dict['identificatie'], BytesIO(b'some content2'))

        self.assertNotEqual(document['titel'], updated_document['titel'])

    def test_update_document_no_stream(self):
        eio = EnkelvoudigInformatieObjectFactory(identificatie='test')
        eio_dict = eio.__dict__

        document = self.backend.create_document(eio_dict.copy(), BytesIO(b'some content'))
        self.assertIsNotNone(document)
        self.assertEqual(document['identificatie'], 'test')

        eio_dict['titel'] = 'test-titel-die-unique-is'

        updated_document = self.backend.update_document(eio_dict.copy(), eio_dict['identificatie'], None)

        self.assertNotEqual(document['titel'], updated_document['titel'])

    # DELETE DOCUMENT TESTS
    def test_delete_documents(self):
        eio = EnkelvoudigInformatieObjectFactory(identificatie='test')
        eio_dict = eio.__dict__

        cmis_doc = cmis_client.create_document(eio.identificatie, eio_dict.copy(), BytesIO(b'some content'))
        self.assertIsNotNone(cmis_doc)
        self.assertFalse(cmis_doc.properties.get('drc:document__verwijderd'))

        self.backend.delete_document('test')
        cmis_doc.reload()
        self.assertTrue(cmis_doc.properties.get('drc:document__verwijderd'))

    # CONNECT DOCUMENT TESTS
    def test_create_case_link(self):
        # Create zaaktype_folder
        zaaktype_folder = cmis_client.get_or_create_zaaktype_folder({
            'url': 'https://ref.tst.vng.cloud/ztc/api/v1/catalogussen/f7afd156-c8f5-4666-b8b5-28a4a9b5dfc7/zaaktypen/0119dd4e-7be9-477e-bccf-75023b1453c1',
            'identificatie': 1,
            'omschrijving': 'Melding Openbare Ruimte',
        })

        # Create zaak_folder
        cmis_client.get_or_create_zaak_folder({
            'url': 'https://ref.tst.vng.cloud/zrc/api/v1/zaken/random-zaak-uuid',
            'identificatie': '1bcfd0d6-c817-428c-a3f4-4047038c184d',
            'zaaktype': 'https://ref.tst.vng.cloud/ztc/api/v1/catalogussen/f7afd156-c8f5-4666-b8b5-28a4a9b5dfc7/zaaktypen/0119dd4e-7be9-477e-bccf-75023b1453c1',
            'startdatum': '2023-12-06',
            'einddatum': None,
            'registratiedatum': '2019-04-17',
            'bronorganisatie': '509381406',
        }, zaaktype_folder)

        # Create document
        eio = EnkelvoudigInformatieObjectFactory(identificatie='test')
        eio_dict = eio.__dict__

        document = self.backend.create_document(eio_dict.copy(), BytesIO(b'some content'))
        self.assertIsNotNone(document)

        self.backend.create_case_link({
            'informatieobject': document['url'],
            'object': 'https://ref.tst.vng.cloud/zrc/api/v1/zaken/random-zaak-uuid'
        })

    def test_create_case_link_without_case_folder(self):
        # Create document
        eio = EnkelvoudigInformatieObjectFactory(identificatie='test')
        eio_dict = eio.__dict__

        document = self.backend.create_document(eio_dict.copy(), None)
        self.assertIsNotNone(document)

        self.backend.create_case_link({
            'informatieobject': document['url'],
            'object': 'https://ref.tst.vng.cloud/zrc/api/v1/zaken/random-zaak-uuid'
        })

    def test_create_case_link_create_copy(self):
        # Create zaaktype_folder
        zaaktype_folder = cmis_client.get_or_create_zaaktype_folder({
            'url': 'https://ref.tst.vng.cloud/ztc/api/v1/catalogussen/f7afd156-c8f5-4666-b8b5-28a4a9b5dfc7/zaaktypen/0119dd4e-7be9-477e-bccf-75023b1453c1',
            'identificatie': 1,
            'omschrijving': 'Melding Openbare Ruimte',
        })

        # Create zaak_folders
        cmis_client.get_or_create_zaak_folder({
            'url': 'https://ref.tst.vng.cloud/zrc/api/v1/zaken/random-zaak-uuid',
            'identificatie': '1bcfd0d6-c817-428c-a3f4-4047038c184d',
            'zaaktype': 'https://ref.tst.vng.cloud/ztc/api/v1/catalogussen/f7afd156-c8f5-4666-b8b5-28a4a9b5dfc7/zaaktypen/0119dd4e-7be9-477e-bccf-75023b1453c1',
            'startdatum': '2023-12-06',
            'einddatum': None,
            'registratiedatum': '2019-04-17',
            'bronorganisatie': '509381406',
        }, zaaktype_folder)

        cmis_client.get_or_create_zaak_folder({
            'url': 'https://ref.tst.vng.cloud/zrc/api/v1/zaken/random-zaak-uuid2',
            'identificatie': '1bcfd0d6-c817-428c-a3f4-4047038c184b',
            'zaaktype': 'https://ref.tst.vng.cloud/ztc/api/v1/catalogussen/f7afd156-c8f5-4666-b8b5-28a4a9b5dfc7/zaaktypen/0119dd4e-7be9-477e-bccf-75023b1453c1',
            'startdatum': '2023-12-06',
            'einddatum': None,
            'registratiedatum': '2019-04-17',
            'bronorganisatie': '509381406',
        }, zaaktype_folder)

        # Create document
        eio = EnkelvoudigInformatieObjectFactory(identificatie='test')
        eio_dict = eio.__dict__

        document = self.backend.create_document(eio_dict.copy(), None)
        self.assertIsNotNone(document)

        self.backend.create_case_link({
            'informatieobject': document['url'],
            'object': 'https://ref.tst.vng.cloud/zrc/api/v1/zaken/random-zaak-uuid'
        })

        self.backend.create_case_link({
            'informatieobject': document['url'],
            'object': 'https://ref.tst.vng.cloud/zrc/api/v1/zaken/random-zaak-uuid2'
        })

    # def test_get_folder(self):
    #     self.assertIsNone(self.backend.get_folder("test"))

    # def test_create_folder(self):
    #     self.assertIsNone(self.backend.create_folder("test"))

    # def test_rename_folder(self):
    #     self.assertIsNone(self.backend.rename_folder("test", "new_test"))

    # def test_remove_folder(self):
    #     self.assertIsNone(self.backend.remove_folder("test"))

    # def test_get_document_without_cmisstorage(self):
    #     eio = EnkelvoudigInformatieObjectFactory()
    #     self.assertIsNone(self.backend.get_document(eio).url)
    #     self.assertIsNone(self.backend.get_document(eio).auteur)
    #     self.assertIsNone(self.backend.get_document(eio).bestandsnaam)
    #     self.assertIsNone(self.backend.get_document(eio).creatiedatum)
    #     self.assertIsNone(self.backend.get_document(eio).vertrouwelijkheidaanduiding)
    #     self.assertIsNone(self.backend.get_document(eio).taal)

    # def test_get_document(self):
    #     enkelvoudiginformatieobject = EnkelvoudigInformatieObjectFactory()
    #     temp_document = self.backend.get_document(enkelvoudiginformatieobject)
    #     print(temp_document.__dict__)
    #     self.assertEqual(
    #         temp_document.url,
    #         f"http://example.com/cmis/content/{enkelvoudiginformatieobject.identificatie}.bin",
    #     )
    #     self.assertEqual(temp_document.auteur, '')
    #     self.assertEqual(temp_document.bestandsnaam, '')
    #     self.assertEqual(temp_document.creatiedatum, '')
    #     self.assertEqual(temp_document.vertrouwelijkheidaanduiding, '')
    #     self.assertEqual(temp_document.taal, '')

    # @override_settings(IS_HTTPS=True)
    # def test_get_document_https(self):
    #     enkelvoudiginformatieobject = EnkelvoudigInformatieObjectFactory()
    #     temp_document = self.backend.get_document(enkelvoudiginformatieobject)
    #     self.assertEqual(
    #         temp_document.url,
    #         f"https://example.com/cmis/content/{enkelvoudiginformatieobject.identificatie}.bin",
    #     )
    #     self.assertEqual(temp_document.auteur, '')
    #     self.assertEqual(temp_document.bestandsnaam, '')
    #     self.assertEqual(temp_document.creatiedatum, '')
    #     self.assertEqual(temp_document.vertrouwelijkheidaanduiding, '')
    #     self.assertEqual(temp_document.taal, '')



from django.utils.translation import ugettext_lazy as _

from djchoices import ChoiceItem, DjangoChoices


class CMISCapabilities(DjangoChoices):
    """
    http://docs.oasis-open.org/cmis/CMIS/v1.0/cmis-spec-v1.0.html
    """

    changes = ChoiceItem(
        "Changes", _('Indicates what level of changes (if any) the repository exposes via the "change log" service.')
    )
    all_versions_searchable = ChoiceItem(
        "AllVersionsSearchable",
        _(
            "Ability of the Repository to include all versions of document. If False, typically either the latest or the latest major version will be searchable."
        ),
    )
    content_stream_updatability = ChoiceItem(
        "ContentStreamUpdatability",
        _("Indicates the support a repository has for updating a document's content stream."),
    )
    pwc_updatable = ChoiceItem(
        "PWCUpdatable", _('Ability for an application to update the "Private Working Copy" of a checked-out document.')
    )
    pwc_searchable = ChoiceItem(
        "PWCSearchable",
        _(
            'Ability of the Repository to include the "Private Working Copy" of checked-out documents in query search scope; otherwise PWC\'s are not searchable'
        ),
    )
    unfiling = ChoiceItem(
        "Unfiling",
        _("Ability for an application to leave a document or other file-able object not filed in any folder."),
    )
    multifiling = ChoiceItem(
        "Multifiling",
        _("Ability for an application to file a document or other file-able object in more than one folder."),
    )
    version_specific_filing = ChoiceItem(
        "VersionSpecificFiling",
        _("Ability for an application to file individual versions (i.e., not all versions) of a document in a folder."),
    )
    renditions = ChoiceItem(
        "Renditions", _("Indicates whether or not the repository exposes renditions of document or folder objects.")
    )
    query = ChoiceItem("Query", _("Indicates the types of queries that the Repository has the ability to fulfill."))
    get_folder_tree = ChoiceItem(
        "GetFolderTree", _("Ability for an application to retrieve the folder tree via the getFolderTree service.")
    )
    acl = ChoiceItem("ACL", _("Indicates the level of support for ACLs by the repository."))
    join = ChoiceItem("Join", _(" Indicates the types of JOIN keywords that the Repository can fulfill in queries."))



"""
Listen to the notifications that are send by the NRC
"""
from vng_api_common.models import APICredential
from vng_api_common.notifications.handlers import default
from zds_client.client import Client

from drc_cmis.client import cmis_client


class ZakenHandler:
    def handle(self, data: dict) -> None:
        if data.get('actie') == 'create' and data.get('resource') == 'zaak':
            zaaktype_url = data.get('kenmerken', {}).get('zaaktype')
            client = Client.from_url(zaaktype_url)
            client.auth = APICredential.get_auth(zaaktype_url)
            client.auth.set_claims(scopes=['zds.scopes.zaaktypes.lezen'])
            zaaktype_data = client.retrieve('zaaktype', url=zaaktype_url)
            zaaktype_folder = cmis_client.get_or_create_zaaktype_folder(zaaktype_data)

            zaak_url = data.get('resource_url')
            client = Client.from_url(zaak_url)
            client.auth = APICredential.get_auth(zaak_url)
            client.auth.set_claims(scopes=[
                'notificaties.scopes.publiceren',
                'zds.scopes.statussen.toevoegen',
                'zds.scopes.zaken.aanmaken',
                'zds.scopes.zaken.bijwerken',
                'zds.scopes.zaken.geforceerd-bijwerken',
                'zds.scopes.zaken.heropenen',
                'zds.scopes.zaken.lezen',
                'zds.scopes.zaken.verwijderen',
            ], zaaktypes=[zaaktype_url])
            zaak_data = client.retrieve('zaak', url=zaak_url)
            cmis_client.get_or_create_zaak_folder(zaak_data, zaaktype_folder)


class RoutingHandler:

    def __init__(self, config: dict, default=None):
        self.config = config
        self.default = default

    def handle(self, message: dict):
        handler = self.config.get(message['kanaal'])
        if handler is not None:
            handler.handle(message)
        elif self.default:
            self.default.handle(message)


default = RoutingHandler({'zaken': ZakenHandler()}, default=default)



import logging

from django.conf import settings
from django.utils.module_loading import import_string

from cmislib.exceptions import UpdateConflictException

from .client import cmis_client
from .convert import create_case_dict_from_cmis_doc, create_dict_from_cmis_doc
from .exceptions import DocumentExistsError

logger = logging.getLogger(__name__)


class CMISDRCStorageBackend(import_string(settings.ABSTRACT_BASE_CLASS)):
    """
    This is the backend that is used to store the documents in a CMIS compatible backend.

    This class is based on:
    drc.backend.abstract.BaseDRCStorageBackend
    """

    # DOCUMENTS ====================================================================================
    # CREATE
    def create_document(self, validated_data, inhoud):
        identificatie = validated_data.pop('identificatie')

        try:
            cmis_doc = cmis_client.create_document(identificatie, validated_data, inhoud)
            dict_doc = create_dict_from_cmis_doc(cmis_doc)
            return dict_doc
        except UpdateConflictException:
            error_message = 'Document is niet uniek. Dit kan liggen aan de titel, inhoud of documentnaam'
            raise self.exception_class({None: error_message}, create=True)
        except DocumentExistsError as e:
            raise self.exception_class({None: e.message}, create=True)

    # READ
    def get_documents(self, filters=None):
        """
        Fetch all the documents from CMIS and create a dict that is compatible with the serializer.
        """
        cmis_documents = cmis_client.get_cmis_documents(filters=filters)
        documents_data = []
        for cmis_doc in cmis_documents:
            dict_document = create_dict_from_cmis_doc(cmis_doc)
            if dict_document:
                documents_data.append(dict_document)

        return documents_data

    def get_document(self, uuid):
        cmis_document = cmis_client.get_cmis_document(identificatie=uuid)
        return create_dict_from_cmis_doc(cmis_document)

    # UPDATE
    def update_document(self, validated_data, identificatie, inhoud):
        cmis_document = cmis_client.update_cmis_document(identificatie, validated_data, inhoud)
        return create_dict_from_cmis_doc(cmis_document)

    # DELETE
    def delete_document(self, uuid):
        cmis_document = cmis_client.delete_cmis_document(identificatie=uuid)
        return create_dict_from_cmis_doc(cmis_document)

    # CONNECTIONS ==================================================================================
    # CREATE
    def create_case_link(self, validated_data):
        """
        There are 2 possible paths here,
        1. There is no case connected to the document yet. So the document can be connected to the case.
        2. There is a case connected to the document, so we need to create a copy of the document.
        """
        document_id = validated_data.get('informatieobject').split('/')[-1]
        cmis_doc = cmis_client.get_cmis_document(identificatie=document_id)
        folder = cmis_client.get_folder_from_case_url(validated_data.get('object'))

        if not cmis_doc.properties.get('drc:connectie__zaakurl'):
            cmis_client.update_case_connection(cmis_doc, validated_data)
            if folder:
                cmis_client.move_to_case(cmis_doc, folder)
            else:
                return
        else:
            cmis_client.copy_document(cmis_doc, folder, validated_data)

    # READ
    def get_document_cases(self):
        """
        Get all documents that have a case url.
        """
        cmis_documents = cmis_client.get_cmis_documents(filter_case=True)
        documents_data = []
        for cmis_doc in cmis_documents:
            dict_document = create_case_dict_from_cmis_doc(cmis_doc)
            if dict_document:
                documents_data.append(dict_document)

        return documents_data

    # UPDATE
    # TODO

    # DELETE
    # TODO



from uuid import uuid4

import factory
from faker import Faker

fake = Faker()


class EnkelvoudigInformatieObjectFactory(factory.DjangoModelFactory):
    identificatie = factory.Sequence(lambda a: uuid4())
    bronorganisatie = factory.Sequence(lambda x: '1234{}'.format(x))
    creatiedatum = factory.Faker('date')
    titel = factory.Faker('word')
    vertrouwelijkheidaanduiding = factory.Faker('word')
    auteur = factory.Faker('first_name')
    formaat = 'some formaat'
    taal = 'dut'
    beschrijving = factory.Faker('paragraph')
    inhoud = factory.django.FileField(data=fake.word().encode('utf-8'), filename=fake.file_name())
    informatieobjecttype = 'https://example.com/ztc/api/v1/catalogus/1/informatieobjecttype/1'

    class Meta:
        model = 'app.EnkelvoudigInformatieObject'



from datetime import date

from cmislib.exceptions import ObjectNotFoundException

from drc_cmis import settings
from drc_cmis.client import cmis_client


class DMSMixin:
    def setUp(self):
        super().setUp()
        today = date.today()
        self.addCleanup(lambda: self._removeTree(f"/{settings.BASE_FOLDER_LOCATION}"))
        cmis_client._base_folder = None

    def _removeTree(self, path):
        try:
            root_folder = cmis_client._repo.getObjectByPath(path)
        except ObjectNotFoundException:
            return
        root_folder.deleteTree()



# from unittest.mock import patch
from django.test import TestCase, override_settings

from drc_cmis.apps import check_cmis


class CheckCMISTests(TestCase):
    def test_check_cmis(self):
        check_cmis('test')

    @override_settings(DRC_CMIS_CLIENT_USER_PASSWORD='Not correct')
    def test_no_correct_password(self):
        check_cmis('test')

    # @patch('drc_cmis.client.CmisClient')
    # def test_check_cmis_no_multifilling(self, client_mock):
    #     client_mock.return_value = 2
    #     check_cmis('test')
    #     print(dir(client_mock))
    #     self.assertTrue(client_mock.called)



import codecs
import logging
from datetime import datetime
from io import BytesIO

from django.utils import timezone

from cmislib import CmisClient
from cmislib.exceptions import UpdateConflictException

from drc_cmis import settings

# from .choices import CMISObjectType
from .exceptions import (
    DocumentConflictException, DocumentDoesNotExistError, DocumentExistsError
)
from .mapper import mapper, reverse_mapper
from .query import CMISQuery

# from .utils import upload_to

logger = logging.getLogger(__name__)


class CMISDRCClient:
    """
    DRC client implementation using the CMIS protocol.
    """
    all_documents_query = CMISQuery("SELECT * FROM cmis:document")
    all_drc_documents_query = CMISQuery("SELECT * FROM drc:document")
    documents_query = CMISQuery("SELECT * FROM drc:document WHERE %s")
    document_cases_query = CMISQuery("SELECT * FROM drc:document WHERE drc:connectie__zaakurl IS NOT NULL")
    document_via_identification_query = CMISQuery("SELECT * FROM drc:document WHERE drc:document__identificatie = '%s'")
    document_query = CMISQuery("SELECT * FROM cmis:document WHERE cmis:objectId = 'workspace://SpacesStore/%s;1.0'")
    find_folder_query = CMISQuery("SELECT * FROM drc:zaakfolder WHERE drc:zaak__url='%s'")

    _repo = None
    _root_folder = None
    _base_folder = None

    @property
    def _get_repo(self):
        """
        Connect to the CMIS repository
        """
        if not self._repo:
            from .models import CMISConfig
            config = CMISConfig.get_solo()
            _client = CmisClient(config.client_url, config.client_user, config.client_password)
            self._repo = _client.getDefaultRepository()
        return self._repo

    @property
    def _get_root_folder(self):
        if not self._root_folder:
            self._root_folder = self._get_repo.getObjectByPath("/")
        return self._root_folder

    @property
    def _get_base_folder(self):
        if not self._base_folder:
            self._base_folder = self._get_or_create_folder(settings.BASE_FOLDER_LOCATION, self._get_root_folder)
        return self._base_folder

    # ZRC Notification client calls.
    def get_or_create_zaaktype_folder(self, zaaktype):
        """
        Create a folder with the prefix 'zaaktype-' to make a zaaktype folder

        TODO: create custom model.
        """
        properties = {
            "cmis:objectTypeId": "F:drc:zaaktypefolder",
            "drc:zaaktype__url": zaaktype.get('url'),
            "drc:zaaktype__identificatie": zaaktype.get('identificatie'),
        }

        folder_name = f"zaaktype-{zaaktype.get('omschrijving')}-{zaaktype.get('identificatie')}"
        cmis_folder = self._get_or_create_folder(folder_name, self._get_base_folder, properties)
        return cmis_folder

    def get_or_create_zaak_folder(self, zaak, zaaktype_folder):
        """
        Create a folder with the prefix 'zaak-' to make a zaak folder
        """

        properties = {
            "cmis:objectTypeId": "F:drc:zaakfolder",
            "drc:zaak__url": zaak.get('url'),
            "drc:zaak__identificatie": zaak.get('identificatie'),
            "drc:zaak__zaaktypeurl": zaak.get('zaaktype'),
            "drc:zaak__bronorganisatie": zaak.get("bronorganisatie"),
        }
        cmis_folder = self._get_or_create_folder(f"zaak-{zaak.get('identificatie')}", zaaktype_folder, properties)
        return cmis_folder

    # DRC client calls.
    def create_document(self, identificatie, data, stream=None):
        """
        :param identificatie: EnkelvoudigInformatieObject identificatie
        :param data: De data voor de properties.
        :param stream: Inhoud van het document.

        :return: AtomPubDocument instance die aangemaakt werd.
        :raises: DocumentExistsError wanneer er al een document met dezelfde identificatie bestaat, binnen de zaakfolder
        """
        self._check_document_exists(identificatie)

        now = datetime.now()

        if stream is None:
            stream = BytesIO()

        year_folder = self._get_or_create_folder(str(now.year), self._get_base_folder)
        month_folder = self._get_or_create_folder(str(now.month), year_folder)
        day_folder = self._get_or_create_folder(str(now.day), month_folder)
        properties = self._build_properties(identificatie, data)

        return self._get_repo.createDocument(
            name=data.get('titel'), properties=properties, contentFile=stream, contentType=None,
            parentFolder=day_folder,
        )

    def get_cmis_documents(self, filter_case=False, filters=None):
        """
        Gives a list of cmis documents.

        :return: :class:`AtomPubDocument` list
        """
        from .models import CMISConfig

        if filters:
            filter_string = ''
            for key, value in filters.items():
                if value:
                    filter_string = f"{mapper(key)} = '{value}' AND"

            if filter_string.endswith(' AND'):
                filter_string = filter_string[:-4]
            query = self.documents_query(filter_string)
        elif filter_case:
            query = self.document_cases_query()
        else:
            config = CMISConfig.get_solo()
            folders = config.locations.values_list('location', flat=True)
            folder_query = ""
            for index, folder in enumerate(folders):
                cmis_folder = self._get_or_create_folder(folder, self._get_root_folder)
                folder_query += f"IN_TREE('{cmis_folder.properties.get('cmis:objectId')}') "
                if index + 1 < len(folders):
                    folder_query += "OR "
            query = self.documents_query(folder_query)

        # Remove the /'s from the string
        query = codecs.escape_decode(query)[0].decode()
        print(query)
        result_set = self._get_repo.query(query)
        unpacked_result_set = [item for item in result_set]
        cmis_documents = [doc.getLatestVersion() for doc in unpacked_result_set]
        return cmis_documents

    def get_cmis_document(self, identificatie, document_query=None):
        """
        Given a cmis document instance.

        :param identificatie.
        :return: :class:`AtomPubDocument` object
        """
        if not document_query:
            document_query = self.document_query
        query = document_query(identificatie)
        result_set = self._get_repo.query(query)
        unpacked_result_set = [item for item in result_set]
        if not unpacked_result_set:
            error_string = "Document met identificatie {} bestaat niet in het CMIS connection".format(identificatie)
            raise DocumentDoesNotExistError(error_string)

        doc = unpacked_result_set[0]
        doc = doc.getLatestVersion()
        return doc

    def update_cmis_document(self, identificatie, data, stream=None):
        cmis_doc = self.get_cmis_document(identificatie, self.document_via_identification_query)

        # build up the properties
        current_properties = cmis_doc.properties
        new_properties = self._build_properties(identificatie, data)
        diff_properties = {
            key: value
            for key, value in new_properties.items()
            if current_properties.get(key) != new_properties.get(key)
        }

        try:
            cmis_doc.updateProperties(diff_properties)
        except UpdateConflictException as exc:
            # Node locked!
            raise DocumentConflictException from exc

        if stream is not None:
            cmis_doc.setContentStream(stream, None)

        return cmis_doc

    def delete_cmis_document(self, identificatie):
        cmis_doc = self.get_cmis_document(identificatie, self.document_via_identification_query)
        new_properties = {'drc:document__verwijderd': True}

        try:
            cmis_doc.updateProperties(new_properties)
        except UpdateConflictException as exc:
            # Node locked!
            raise DocumentConflictException from exc
        return cmis_doc

    def update_case_connection(self, cmis_doc, data):
        current_properties = cmis_doc.properties
        new_properties = self._build_case_properties(data)
        diff_properties = {
            key: value
            for key, value in new_properties.items()
            if current_properties.get(key) != new_properties.get(key)
        }

        try:
            cmis_doc.updateProperties(diff_properties)
        except UpdateConflictException as exc:
            # Node locked!
            raise DocumentConflictException from exc

    def move_to_case(self, cmis_doc, folder):
        parent = [parent for parent in cmis_doc.getObjectParents()][0]
        cmis_doc.move(parent, folder)

    def copy_document(self, cmis_doc, folder, data):
        properties = cmis_doc.properties
        new_properties = self._build_case_properties(data)
        properties.update(**new_properties)
        print(properties)
        # TODO: Make this an actual copy function.

        return cmis_client._repo.createDocumentFromSource(cmis_doc.getObjectId(), folder, properties)

    def get_folder_from_case_url(self, zaak_url):
        query = self.find_folder_query(zaak_url)
        result_set = self._get_repo.query(query)
        unpacked_result_set = [item for item in result_set.getResults()]
        if unpacked_result_set:
            folder = unpacked_result_set[0]
            # ! A reload is done because there are some important values missing from the objects.
            folder.reload()
            return folder
        return None

    # Private functions.
    # TODO: Paste private functions.
    def _get_or_create_folder(self, name, parent, properties=None):
        """
        Get or create the folder with :param:`name` in :param:`parent`.

        :param name: string, the name of the folder to create.
        :param parent: parent folder to create the folder in as subfolder.
        :param properties: dictionary with cmis and/or custom properties to
          pass to the folder object
        :return: the folder that was retrieved or created.
        """
        existing = self._get_folder(name, parent)
        if existing:
            return existing
        return parent.createFolder(name, properties=properties or {})

    def _get_folder(self, name, parent):
        existing = next((child for child in parent.getChildren() if str(child.name) == str(name)), None)
        if existing is not None:
            return existing
        return None

    def _build_properties(self, identificatie, data):
        base_properties = {mapper(key): value for key, value in data.items() if mapper(key)}
        base_properties["cmis:objectTypeId"] = "D:drc:document"
        base_properties[mapper("identificatie")] = identificatie
        return base_properties
        # {
        #     "cmis:objectTypeId": 'D:drc:document',  # Set the type of document that is uploaded.
        #     "cmis:name": data.get('titel'),
        #     "drc:document__identificatie": identificatie,
        #     "drc:document__bronorganisatie": data.get('bronorganisatie'),
        #     "drc:document__creatiedatum": data.get('creatiedatum'),
        #     "drc:document__vertrouwelijkaanduiding": data.get('vertrouwelijkaanduiding', ''),
        #     "drc:document__auteur": data.get('auteur'),
        #     "drc:document__status": data.get('status', ''),
        #     "drc:document__beschrijving": data.get('beschrijving', ''),
        #     "drc:document__ontvangstdatum": data.get('ontvangstdatum'),
        #     "drc:document__verzenddatum": data.get('verzenddatum'),
        #     "drc:document__indicatiegebruiksrecht": data.get('indicatie_gebruiksrecht', ''),
        #     "drc:document__ondertekeningsoort": data.get('ondertekening_soort', ''),
        #     "drc:document__ondertekeningdatum": data.get('ondertekening_datum'),
        #     "drc:document__informatieobjecttype": data.get('informatieobjecttype', ''),
        #     "drc:document__formaat": data.get('formaat', ''),
        #     "drc:document__taal": data.get('taal'),
        #     "drc:document__bestandsnaam": data.get('bestandsnaam', ''),
        #     "drc:document__link": data.get('link', ''),
        #     "drc:document__integriteitalgoritme": data.get('integriteit_algoritme', ''),
        #     "drc:document__integriteitwaarde": data.get('integriteit_waarde', ''),
        #     "drc:document__integriteitdatum": data.get('integriteit_datum'),
        # }

    def _build_case_properties(self, data):
        return {
            "drc:connectie__zaakurl": data.get('object'),
            "drc:connectie__objecttype": data.get('objectType'),
            "drc:connectie__aardrelatieweergave": data.get('aardRelatieWeergave'),
            "drc:connectie__titel": data.get('titel'),
            "drc:connectie__beschrijving": data.get('beschrijving'),
            "drc:connectie__registratiedatum": timezone.now().date(),
        }

    def _check_document_exists(self, identificatie):
        try:
            self.get_cmis_document(identificatie, self.document_via_identification_query)
        except DocumentDoesNotExistError:
            pass
        else:
            error_string = "Document identificatie {} is niet uniek".format(identificatie)
            raise DocumentExistsError(error_string)

    # ! Not used yet.
    # ! Fix
    # def get_folder_name(self, zaak_url, folder_config):
    #     name = ""
    #     if folder_config.type == CMISObjectType.zaak_folder:
    #         name = slugify(zaak_url)
    #     else:
    #         if not folder_config.name:
    #             raise ValueError(("Could not determine a folder name for zaak {}").format(slugify(zaak_url)))
    #     return folder_config.name or name

    # def _get_zaakfolder(self, zaak_url):
    #     bits = [self.get_folder_name(zaak_url, folder_config) for folder_config in upload_to(zaak_url)]
    #     path = "/" + ("/").join(bits)
    #     return self._get_repo.getObjectByPath(path)

    # def _build_cmis_doc_properties(self, connection, filename=None):
    #     properties = connection.get_cmis_properties()
    #     properties["cmis:objectTypeId"] = CMISObjectType.edc
    #     if filename is not None:
    #         properties["cmis:name"] = filename
    #     return properties

    # def creeer_zaakfolder(self, zaak_url):
    #     """
    #     Maak de zaak folder aan in het DRC.

    #     :param zaak_url: Een link naar de zaak.
    #     :return: :class:`cmslib.atompub_binding.AtomPubFolder` object - de
    #       cmslib representatie van de (aangemaakte) zaakmap.
    #     """
    #     upload_to_folder = upload_to(zaak_url)
    #     for folder_config in upload_to_folder:
    #         if folder_config.type == CMISObjectType.zaken:
    #             folder_config.type = "cmis:folder"

    #     parent = None
    #     for folder_config in upload_to_folder:
    #         properties = {"cmis:objectTypeId": folder_config.type} if folder_config.type else {}
    #         name = self.get_folder_name(zaak_url, folder_config)
    #         parent, _ = self._get_or_create_folder(name, properties, parent=parent)

    #     zaak_folder = parent
    #     return zaak_folder

    # def geef_inhoud(self, document):
    #     """
    #     Retrieve the document via its identifier from the DRC.

    #     :param document: EnkelvoudigInformatieObject instance
    #     :return: tuple of (filename, BytesIO()) with the stream filename and the binary content
    #     """
    #     try:
    #         doc = self.get_cmis_document(document)
    #     except DocumentDoesNotExistError:
    #         return (None, BytesIO())

    #     filename = doc.properties["cmis:name"]
    #     empty = doc.properties["cmis:contentStreamId"] is None
    #     if empty:
    #         return (filename, BytesIO())
    #     return (filename, doc.getContentStream())

    # def relateer_aan_zaak(self, document, zaak_url):
    #     """
    #     Wijs het document aan :param:`zaak` toe.

    #     Verplaatst het document van de huidige folder naar de zaakfolder.
    #     """
    #     cmis_doc = self.get_cmis_document(document)
    #     zaakfolder = self._get_zaakfolder(zaak_url)
    #     parent = [parent for parent in cmis_doc.getObjectParents()][0]
    #     cmis_doc.move(parent, zaakfolder)

    # def checkout(self, document):
    #     """
    #     Checkout (lock) the requested document and return the PWC ID + check out username.

    #     :param document: :class:`EnkelvoudigInformatieObject` instance.
    #     """
    #     cmis_doc = self.get_cmis_document(document)
    #     try:
    #         pwc = cmis_doc.checkout()
    #     except UpdateConflictException as exc:
    #         raise DocumentLockedException("Document was already checked out") from exc

    #     pwc.reload()
    #     checkout_id = pwc.properties["cmis:versionSeriesCheckedOutId"]
    #     checkout_by = pwc.properties["cmis:versionSeriesCheckedOutBy"]
    #     return (checkout_id, checkout_by)

    # def cancel_checkout(self, document, checkout_id):
    #     cmis_doc = self.get_cmis_document(document, checkout_id=checkout_id)
    #     cmis_doc.cancelCheckout()

    # def ontkoppel_zaakdocument(self, document, zaak_url):
    #     cmis_doc = self.get_cmis_document(document)
    #     cmis_folder = self._get_zaakfolder(zaak_url)
    #     trash_folder, _ = self._get_or_create_folder(self.TRASH_FOLDER)
    #     cmis_doc.move(cmis_folder, trash_folder)

    # def gooi_in_prullenbak(self, document):
    #     cmis_doc = self.get_cmis_document(document)
    #     trash_folder, _ = self._get_or_create_folder(self.TRASH_FOLDER)
    #     default_folder, _ = self._get_or_create_folder(self.TEMP_FOLDER_NAME)
    #     cmis_doc.move(default_folder, trash_folder)

    # def is_locked(self, document):
    #     cmis_doc = self.get_cmis_document(document)
    #     pwc = cmis_doc.getPrivateWorkingCopy()
    #     return pwc is not None

    # def verwijder_document(self, document):
    #     cmis_doc = self.get_cmis_document(document)
    #     cmis_doc.delete()


cmis_client = CMISDRCClient()
